import sys
import os

cpu_index = 0
cpu_total = 1
job_index = 0

if len(sys.argv)>=3:
	cpu_index = int(sys.argv[1])
	cpu_total = int(sys.argv[2])
else:
	print("warning : arguments might be missing")

def add_job(func):
	global job_index
	if job_index == cpu_index:
		print(f"I'm running job {job_index}")
		func()
	job_index = (job_index+1)%cpu_total

###############################################################

from functools import reduce # Valid in Python 2.6+, required in Python 3
import operator

import torch
import torch.nn as nn
import torch.nn.functional as F

import anyact as aa

import numpy as np
np.seterr(divide='ignore', invalid='ignore')

import torchvision
import torchvision.datasets as datasets

import os
import h5py

from toy_problem_mnist_analysis import analysis_classification
from toy_problem_mnist import fmnist,mnist,mnist_perf

import matplotlib.pyplot as plt

device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

def expspace(x0,x1,N): return torch.exp(torch.linspace(np.log(x0),np.log(x1),N))

################################################################
################################################################
################################################################
################################################################

def nn_fc(layers,σ,dropout=0.5,batchnorm=1,gen_linear=lambda dim_i,dim_o: nn.Linear(dim_i,dim_o)):
	layers = [int(np.round(x)) for x in layers]
	disable_drp = 0 if dropout==0 else 1
	disable_btn = batchnorm
	if hasattr(σ,"m"): m=σ.m
	else: m = 1
	return sum([
		[gen_linear((m if not index==0 else 1)*λ1,λ2),σ,
		]+disable_btn*[nn.BatchNorm1d(λ2*m),
		]+disable_drp*[aa.Dropout(dropout),
		]
		for index,(λ1,λ2) in enumerate(zip(layers[0:-2],layers[1:-1]))
	],[])+[gen_linear(m*layers[-2],layers[-1])] # remove last GELU, BatchNorm and Dropout

def bernoulli(shape,p): return torch.bernoulli(torch.ones(shape)*p)

def random_sign_flip(x,proba):
	"""
		proba=0   : never flip
		proba=0.5 : complete random sign
		proba=1.0 : always flip
	"""
	flip = bernoulli(x.shape,proba).to(x.device)
	return x*(1.0-2.0*flip)

def only_sign(x): return 1.0-2.0*(x<0)

from gradiator import gradiator_1
from gradiator import gradiator_2
from gradiator import gradiator_3



# class LinearBroken(nn.Module):
# 	def __init__(self,dim_i,dim_o,
# 		θ_noise_static_σ=0.0,
# 		θ_noise_static_μ=0.0,
# 		# gθ_noise_static_σ=0.0,
# 		# gθ_noise_static_μ=0.0,
# 		# gx_noise_static_σ=0.0,
# 		# gx_noise_static_μ=0.0,
# 		gθ_flip_static=0.0,
# 		gx_flip_static=0.0,
# 		gθ_flip_dynamic=0.0, # 0.48
# 		gx_flip_dynamic=0.0, # 0.48
# 		# θ_only_sign=True ,θ_noise_dynamic=0.00, θ_flip=0.3,
# 		gθ_only_sign=False,
# 		gx_only_sign=False,
# 	):
# 		super().__init__()
# 		self.θ = nn.Parameter(nn.init.xavier_normal_(torch.empty(dim_i, dim_o)))
# 		self.b = nn.Parameter(torch.randn(dim_o))
# 		self.θ_noise_static  = nn.parameter.Parameter( θ_noise_static_μ +  θ_noise_static_σ*torch.randn(dim_i,dim_o),False)
# 		# self.gθ_noise_static = nn.parameter.Parameter(gθ_noise_static_μ + gθ_noise_static_σ*torch.randn(dim_i,dim_o),False)
# 		# self.gx_noise_static = nn.parameter.Parameter(gx_noise_static_μ + gx_noise_static_σ*torch.randn(dim_i      ),False)
# 		self.gθ_flip_static = nn.parameter.Parameter( random_sign_flip(torch.ones(dim_i,dim_o),gθ_flip_static), False)
# 		self.gx_flip_static = nn.parameter.Parameter( random_sign_flip(torch.ones(dim_i      ),gx_flip_static), False)
# 		def fwd(x,θ): return torch.einsum('ij,ni->nj', θ, x) + self.b
# 		def bwd(x,θ,y,g_o,debug=False):
# 			############## break θ
# 			# print(torch.sum(torch.abs(θ)))
# 			# print(torch.sum(torch.abs(self.θ_noise_static)))
# 			θ = θ + self.θ_noise_static
# 			# if θ_only_sign: θ = only_sign(θ)
# 			# θ = random_sign_flip(θ, θ_flip)
# 			############## normal stuff
# 			g_iθ = torch.einsum('ni,nj->nij', x, g_o) # produit extérieur
# 			g_ix = torch.einsum('nj,ij->ni' , g_o, θ) # θ⋅x
# 			############## static breaking
# 			# g_iθ = g_iθ + self.gθ_noise_static
# 			# g_ix = g_ix + self.gx_noise_static
# 			g_iθ = g_iθ * self.gθ_flip_static
# 			g_ix = g_ix * self.gx_flip_static
# 			############## dynamic breaking
# 			if gx_only_sign: g_ix = only_sign(g_ix) # this seems fine...
# 			if gθ_only_sign: g_iθ = only_sign(g_iθ) # this is bad !
# 			if not gx_flip_dynamic==0.0 : g_ix = random_sign_flip(g_ix, gx_flip_dynamic)  # this is
# 			if not gθ_flip_dynamic==0.0 : g_iθ = random_sign_flip(g_iθ, gθ_flip_dynamic)  # but this is ok ... ??
# 			############## output
# 			return g_ix, g_iθ
# 		self.model = gradiator_2(fwd,bwd)
# 	def forward(self,x): return self.model(x,self.θ)

# class LinearBroken(nn.Module):
# 	def __init__(self,dim_i,dim_o,
# 		θ_noise_static=0.0,
# 		θ_noise_dyn   =0.0,
# 		θ_only_sign   =False,
# 		# θ_flip=0.3,
# 		# gθ_noise_dyn_σ=0.0,
# 		# gx_noise_dyn_σ=0.0,
# 		gθ_flip_static=0.0,
# 		gx_flip_static=0.0,
# 		gθ_flip_dynamic=0.0, # 0.48
# 		gx_flip_dynamic=0.0, # 0.48
# 		gθ_only_sign=False,
# 		gx_only_sign=False,
# 	):
# 		super().__init__()
# 		self.θ = nn.Parameter(nn.init.xavier_normal_(torch.empty(dim_i, dim_o)))
# 		self.b = nn.Parameter(torch.randn(dim_o))
# 		self.θ_noise_static  = nn.parameter.Parameter( θ_noise_static*torch.randn(dim_i,dim_o),False)
# 		self.gθ_flip_static = nn.parameter.Parameter( random_sign_flip(torch.ones(dim_i,dim_o),gθ_flip_static), False)
# 		self.gx_flip_static = nn.parameter.Parameter( random_sign_flip(torch.ones(dim_i      ),gx_flip_static), False)
# 		def fwd(x,θ): return torch.einsum('ij,ni->nj', θ, x) + self.b
# 		def bwd(x,θ,y,g_o,debug=False):
# 			############## break θ
# 			θ = θ + self.θ_noise_static
# 			if not θ_noise_dyn==0 : θ = θ + θ_noise_dyn*torch.randn(dim_i,dim_o)
# 			if θ_only_sign: θ = only_sign(θ)
# 			# θ = random_sign_flip(θ, θ_flip)
# 			############## normal stuff
# 			g_iθ = torch.einsum('ni,nj->nij', x, g_o) # produit extérieur
# 			g_ix = torch.einsum('nj,ij->ni' , g_o, θ) # θ⋅x
# 			############## static breaking
# 			# g_iθ = g_iθ + self.gθ_noise_static
# 			# g_ix = g_ix + self.gx_noise_static
# 			g_iθ = g_iθ * self.gθ_flip_static
# 			g_ix = g_ix * self.gx_flip_static
# 			############## dynamic breaking
# 			if gx_only_sign: g_ix = only_sign(g_ix) # this seems fine...
# 			if gθ_only_sign: g_iθ = only_sign(g_iθ) # this is bad !
# 			if not gx_flip_dynamic==0.0 : g_ix = random_sign_flip(g_ix, gx_flip_dynamic)  # this is
# 			if not gθ_flip_dynamic==0.0 : g_iθ = random_sign_flip(g_iθ, gθ_flip_dynamic)  # but this is ok ... ??
# 			############## output
# 			return g_ix, g_iθ
# 		self.model = gradiator_2(fwd,bwd)
# 	def forward(self,x): return self.model(x,self.θ)


class LinearBroken(nn.Module):
	def __init__(self,dim_i,dim_o,
		θ_noise_static=0.0,
		θ_noise_dyn   =0.0,
		θ_only_sign   =False,
		θ_flip        =0.0, # 0.3,
		gθ_flip_static=0.0,
		gx_flip_static=0.0,
		# gθ_noise_dyn_σ=0.0,
		# gx_noise_dyn_σ=0.0,
		gθ_flip_dynamic=0.0, # 0.48
		gx_flip_dynamic=0.0, # 0.48
		gθ_only_sign=False,
		gx_only_sign=False,
	):
		super().__init__()
		self.θ = nn.Parameter(nn.init.xavier_normal_(torch.empty(dim_i, dim_o)))
		self.b = nn.Parameter(torch.randn(dim_o))
		self.θ_noise_static  = nn.parameter.Parameter( θ_noise_static*torch.randn(dim_i,dim_o),False)
		self.gθ_flip_static = nn.parameter.Parameter( random_sign_flip(torch.ones(dim_i,dim_o),gθ_flip_static), False)
		self.gx_flip_static = nn.parameter.Parameter( random_sign_flip(torch.ones(dim_i      ),gx_flip_static), False)
		def fwd(x,θ): return torch.einsum('ij,ni->nj', θ, x) + self.b
		def bwd(x,θ,y,g_o,debug=False):
			############## break θ
			θ = θ + self.θ_noise_static
			if not θ_noise_dyn==0 : θ = θ + θ_noise_dyn*torch.randn(dim_i,dim_o)
			if θ_only_sign: θ = only_sign(θ)
			# θ = random_sign_flip(θ, θ_flip)
			############## normal stuff
			g_iθ = torch.einsum('ni,nj->nij', x, g_o) # produit extérieur
			g_ix = torch.einsum('nj,ij->ni' , g_o, θ) # θ⋅x
			############## flip
			g_iθ = g_iθ * self.gθ_flip_static
			g_ix = g_ix * self.gx_flip_static
			if not gx_flip_dynamic==0.0 : g_ix = random_sign_flip(g_ix, gx_flip_dynamic)  # this is
			if not gθ_flip_dynamic==0.0 : g_iθ = random_sign_flip(g_iθ, gθ_flip_dynamic)  # but this is ok ... ??
			############## only sign
			if gx_only_sign: g_ix = only_sign(g_ix) # this seems fine...
			if gθ_only_sign: g_iθ = only_sign(g_iθ) # this is bad !
			############## output
			return g_ix, g_iθ
		self.model = gradiator_2(fwd,bwd)
	def forward(self,x): return self.model(x,self.θ)


def helper(
	N=4,
	θ_noise_static=0.0,
	θ_noise_dyn   =0.0,
	θ_only_sign   =False,
	θ_flip        =0.0, # 0.3,
	gθ_flip_static=0.0,
	gx_flip_static=0.0,
	# gθ_noise_dyn_σ=0.0,
	# gx_noise_dyn_σ=0.0,
	gθ_flip_dynamic=0.0, # 0.48
	gx_flip_dynamic=0.0, # 0.48
	gθ_only_sign=False,
	gx_only_sign=False,
):
	identifier = f"fc({N})-θnoise=[{θ_noise_static},{θ_noise_dyn}]-gθflip=[{gθ_flip_static},{gθ_flip_dynamic}]"
	identifier = f"{identifier}-gxflip=[{gx_flip_static},{gx_flip_dynamic}]-sgn_only=[{gx_only_sign},{gθ_only_sign}]"
	classifier = nn.Sequential(
		nn.Flatten(),
		*nn_fc(expspace(28*28,10,N+1),aa.modulize(lambda x:x**2),dropout=0,batchnorm=1,gen_linear=lambda dim_i,dim_o:LinearBroken(dim_i,dim_o,
			θ_noise_static=0.0,
			θ_noise_dyn   =0.0,
			θ_only_sign   =False,
			θ_flip        =0.0, # 0.3,
			gθ_flip_static=0.0,
			gx_flip_static=0.0,
			# gθ_noise_dyn_σ=0.0,
			# gx_noise_dyn_σ=0.0,
			gθ_flip_dynamic=0.0, # 0.48
			gx_flip_dynamic=0.0, # 0.48
			gθ_only_sign=False,
			gx_only_sign=False,
		)),
		nn.Softmax(dim=1)
	)
	analysis_classification(classifier, identifier, nn.MSELoss(),)






helper(
	N=4,
	θ_noise_static=0.0,
	θ_noise_dyn   =0.0,
	θ_only_sign   =False,
	θ_flip        =0.0, # 0.3,
	gθ_flip_static=0.0,
	gx_flip_static=0.0,
	# gθ_noise_dyn_σ=0.0,
	# gx_noise_dyn_σ=0.0,
	gθ_flip_dynamic=0.0, # 0.48
	gx_flip_dynamic=0.0, # 0.48
	gθ_only_sign=False,
	gx_only_sign=False,
)













































################################################################
################################################################
################################################################
################################################################
################################################################
################################################################
################################################################
################################################################

for N in [4,8]:
	identifier = f"fc({N})-LinearNormal"
	classifier = nn.Sequential(
		nn.Flatten(),
		*nn_fc(expspace(28*28,10,N+1),nn.GELU(),dropout=0,batchnorm=1),
		nn.Softmax(dim=1)
	)
	analysis_classification(classifier, identifier, nn.MSELoss(),)

################################################################
################################################################
################################################################
################################################################
################################################################
################################################################
################################################################
################################################################

class LinearRandomBwd(nn.Module):
	def __init__(self,dim_i,dim_o):
		super().__init__()
		self.θ = nn.Parameter(nn.init.xavier_normal_(torch.empty(dim_i, dim_o)))
		self.b = nn.Parameter(torch.randn(dim_o))
		def fwd(x,θ): return torch.einsum('ij,ni->nj', θ, x) + self.b
		def bwd(x,θ,y,g_o,debug=False):
			############## break θ
			θ = torch.randn(dim_i,dim_o,device=θ.device)
			############## normal stuff
			g_iθ = torch.einsum('ni,nj->nij', x, g_o) # produit extérieur
			g_ix = torch.einsum('nj,ij->ni' , g_o, θ) # θ⋅x
			############## output
			return g_ix, g_iθ
		self.model = gradiator_2(fwd,bwd)
	def forward(self,x): return self.model(x,self.θ)

for N in [4,8]:
# for N in [8]:
	identifier = f"fc({N})-LinearRandomBwd"
	classifier = nn.Sequential(
		nn.Flatten(),
		*nn_fc(expspace(28*28,10,N+1),nn.GELU(),dropout=0,batchnorm=1,gen_linear=lambda dim_i,dim_o:LinearRandomBwd(dim_i,dim_o,)),
		nn.Softmax(dim=1)
	)
	analysis_classification(classifier, identifier, nn.MSELoss(),)

################################################################
################################################################
################################################################
################################################################
################################################################
################################################################

class LinearθSign(nn.Module):
	def __init__(self,dim_i,dim_o):
		super().__init__()
		self.θ = nn.Parameter(nn.init.xavier_normal_(torch.empty(dim_i, dim_o)))
		self.b = nn.Parameter(torch.randn(dim_o))
		def fwd(x,θ): return torch.einsum('ij,ni->nj', θ, x) + self.b
		def bwd(x,θ,y,g_o,debug=False):
			############## break θ
			θ = only_sign(θ)
			############## normal stuff
			g_iθ = torch.einsum('ni,nj->nij', x, g_o) # produit extérieur
			g_ix = torch.einsum('nj,ij->ni' , g_o, θ) # θ⋅x
			############## output
			return g_ix, g_iθ
		self.model = gradiator_2(fwd,bwd)
	def forward(self,x): return self.model(x,self.θ)

for N in [4,8]:
# for N in [8]:
	identifier = f"fc({N})-LinearθSign"
	classifier = nn.Sequential(
		nn.Flatten(),
		*nn_fc(expspace(28*28,10,N+1),nn.GELU(),dropout=0,batchnorm=1,gen_linear=lambda dim_i,dim_o:LinearθSign(dim_i,dim_o,)),
		nn.Softmax(dim=1)
	)
	analysis_classification(classifier, identifier, nn.MSELoss(),)

################################################################
################################################################
################################################################
################################################################
################################################################
################################################################

class LinearXθSign(nn.Module):
	def __init__(self,dim_i,dim_o,gx_only_sign,gθ_only_sign):
		super().__init__()
		self.θ = nn.Parameter(nn.init.xavier_normal_(torch.empty(dim_i, dim_o)))
		self.b = nn.Parameter(torch.randn(dim_o))
		def fwd(x,θ): return torch.einsum('ij,ni->nj', θ, x) + self.b
		def bwd(x,θ,y,g_o,debug=False):
			############## normal stuff
			g_iθ = torch.einsum('ni,nj->nij', x, g_o) # produit extérieur
			g_ix = torch.einsum('nj,ij->ni' , g_o, θ) # θ⋅x
			############## alter
			if gx_only_sign: g_ix = only_sign(g_ix)
			if gθ_only_sign: g_iθ = only_sign(g_iθ)
			############## output
			return g_ix, g_iθ
		self.model = gradiator_2(fwd,bwd)
	def forward(self,x): return self.model(x,self.θ)

for N in [4,8]:
# for N in [8]:
	identifier = f"fc({N})-LinearGxSign"
	classifier = nn.Sequential(
		nn.Flatten(),
		*nn_fc(expspace(28*28,10,N+1),nn.GELU(),dropout=0,batchnorm=1,gen_linear=lambda dim_i,dim_o:LinearXθSign(dim_i,dim_o,True,False)),
		nn.Softmax(dim=1)
	)
	analysis_classification(classifier, identifier, nn.MSELoss(),)
	identifier = f"fc({N})-LinearGθSign"
	classifier = nn.Sequential(
		nn.Flatten(),
		*nn_fc(expspace(28*28,10,N+1),nn.GELU(),dropout=0,batchnorm=1,gen_linear=lambda dim_i,dim_o:LinearXθSign(dim_i,dim_o,False,True)),
		nn.Softmax(dim=1)
	)
	analysis_classification(classifier, identifier, nn.MSELoss(),)

################################################################
################################################################
################################################################
################################################################
################################################################
################################################################
class LinearXθSignSMALL(nn.Module):
	def __init__(self,dim_i,dim_o,gx_only_sign,gθ_only_sign):
		super().__init__()
		self.θ = nn.Parameter(nn.init.xavier_normal_(torch.empty(dim_i, dim_o)))
		self.b = nn.Parameter(torch.randn(dim_o))
		def fwd(x,θ): return torch.einsum('ij,ni->nj', θ, x) + self.b
		def bwd(x,θ,y,g_o,debug=False):
			############## normal stuff
			g_iθ = torch.einsum('ni,nj->nij', x, g_o) # produit extérieur
			g_ix = torch.einsum('nj,ij->ni' , g_o, θ) # θ⋅x
			############## alter
			if gx_only_sign: g_ix = 0.01*only_sign(g_ix)
			if gθ_only_sign: g_iθ = 0.01*only_sign(g_iθ)
			############## output
			return g_ix, g_iθ
		self.model = gradiator_2(fwd,bwd)
	def forward(self,x): return self.model(x,self.θ)

# for N in [4,8]:
# for N in [8]:
for N in [4]:
	identifier = f"fc({N})-LinearGxSignSMALL2"
	classifier = nn.Sequential(
		nn.Flatten(),
		*nn_fc(expspace(28*28,10,N+1),nn.GELU(),dropout=0,batchnorm=1,gen_linear=lambda dim_i,dim_o:LinearXθSignSMALL(dim_i,dim_o,True,False)),
		nn.Softmax(dim=1)
	)
	# analysis_classification(classifier, identifier, nn.MSELoss(),)
	analysis_classification(classifier, identifier, nn.MSELoss(),epochs=512,batch_size=64)
	# 
	identifier = f"fc({N})-LinearGθSignSMALL2"
	classifier = nn.Sequential(
		nn.Flatten(),
		*nn_fc(expspace(28*28,10,N+1),nn.GELU(),dropout=0,batchnorm=1,gen_linear=lambda dim_i,dim_o:LinearXθSignSMALL(dim_i,dim_o,False,True)),
		nn.Softmax(dim=1)
	)
	# analysis_classification(classifier, identifier, nn.MSELoss(),)
	analysis_classification(classifier, identifier, nn.MSELoss(),epochs=512,batch_size=64)

################################################################
################################################################
################################################################
################################################################
################################################################
################################################################
# class LinearXθSignTINY(nn.Module):
# 	def __init__(self,dim_i,dim_o,gx_only_sign,gθ_only_sign):
# 		super().__init__()
# 		self.θ = nn.Parameter(nn.init.xavier_normal_(torch.empty(dim_i, dim_o)))
# 		self.b = nn.Parameter(torch.randn(dim_o))
# 		def fwd(x,θ): return torch.einsum('ij,ni->nj', θ, x) + self.b
# 		def bwd(x,θ,y,g_o,debug=False):
# 			############## normal stuff
# 			g_iθ = torch.einsum('ni,nj->nij', x, g_o) # produit extérieur
# 			g_ix = torch.einsum('nj,ij->ni' , g_o, θ) # θ⋅x
# 			############## alter
# 			if gx_only_sign: g_ix = 0.0001*only_sign(g_ix)
# 			if gθ_only_sign: g_iθ = 0.0001*only_sign(g_iθ)
# 			############## output
# 			return g_ix, g_iθ
# 		self.model = gradiator_2(fwd,bwd)
# 	def forward(self,x): return self.model(x,self.θ)

# for N in [4,8]:
# # for N in [8]:
# 	identifier = f"fc({N})-LinearGxSignTINY"
# 	classifier = nn.Sequential(
# 		nn.Flatten(),
# 		*nn_fc(expspace(28*28,10,N+1),nn.GELU(),dropout=0,batchnorm=1,gen_linear=lambda dim_i,dim_o:LinearXθSignTINY(dim_i,dim_o,True,False)),
# 		nn.Softmax(dim=1)
# 	)
# 	analysis_classification(classifier, identifier, nn.MSELoss(),)
# 	identifier = f"fc({N})-LinearGθSignTINY"
# 	classifier = nn.Sequential(
# 		nn.Flatten(),
# 		*nn_fc(expspace(28*28,10,N+1),nn.GELU(),dropout=0,batchnorm=1,gen_linear=lambda dim_i,dim_o:LinearXθSignTINY(dim_i,dim_o,False,True)),
# 		nn.Softmax(dim=1)
# 	)
# 	analysis_classification(classifier, identifier, nn.MSELoss(),epochs=512,batch_size=64)

################################################################
################################################################
################################################################
################################################################
################################################################
################################################################

class LinearXθNoiseAdd(nn.Module):
	def __init__(self,dim_i,dim_o,gx_noise,gθ_noise):
		super().__init__()
		self.θ = nn.Parameter(nn.init.xavier_normal_(torch.empty(dim_i, dim_o)))
		self.b = nn.Parameter(torch.randn(dim_o))
		def fwd(x,θ): return torch.einsum('ij,ni->nj', θ, x) + self.b
		def bwd(x,θ,y,g_o,debug=False):
			############## normal stuff
			g_iθ = torch.einsum('ni,nj->nij', x, g_o) # produit extérieur
			g_ix = torch.einsum('nj,ij->ni' , g_o, θ) # θ⋅x
			############## alter
			if gx_noise: g_ix = g_ix + torch.randn(g_ix.shape,device=g_ix.device) # this seems fine...
			if gθ_noise: g_iθ = g_iθ + torch.randn(g_iθ.shape,device=g_iθ.device) # this is bad !
			############## output
			return g_ix, g_iθ
		self.model = gradiator_2(fwd,bwd)
	def forward(self,x): return self.model(x,self.θ)

for N in [4,8]:
	identifier = f"fc({N})-LinearGxNoiseAdd"
	classifier = nn.Sequential(
		nn.Flatten(),
		*nn_fc(expspace(28*28,10,N+1),nn.GELU(),dropout=0,batchnorm=1,gen_linear=lambda dim_i,dim_o:LinearXθNoiseAdd(dim_i,dim_o,True,False)),
		nn.Softmax(dim=1)
	)
	analysis_classification(classifier, identifier, nn.MSELoss(),)
	identifier = f"fc({N})-LinearGθNoiseAdd"
	classifier = nn.Sequential(
		nn.Flatten(),
		*nn_fc(expspace(28*28,10,N+1),nn.GELU(),dropout=0,batchnorm=1,gen_linear=lambda dim_i,dim_o:LinearXθNoiseAdd(dim_i,dim_o,False,True)),
		nn.Softmax(dim=1)
	)
	analysis_classification(classifier, identifier, nn.MSELoss(),)

################################################################
################################################################
################################################################
################################################################
################################################################
################################################################

class LinearXθNoiseMul(nn.Module):
	def __init__(self,dim_i,dim_o,gx_noise,gθ_noise):
		super().__init__()
		self.θ = nn.Parameter(nn.init.xavier_normal_(torch.empty(dim_i, dim_o)))
		self.b = nn.Parameter(torch.randn(dim_o))
		def fwd(x,θ): return torch.einsum('ij,ni->nj', θ, x) + self.b
		def bwd(x,θ,y,g_o,debug=False):
			############## normal stuff
			g_iθ = torch.einsum('ni,nj->nij', x, g_o) # produit extérieur
			g_ix = torch.einsum('nj,ij->ni' , g_o, θ) # θ⋅x
			############## alter
			if gx_noise: g_ix = g_ix * torch.randn(g_ix.shape,device=g_ix.device) # this seems fine...
			if gθ_noise: g_iθ = g_iθ * torch.randn(g_iθ.shape,device=g_iθ.device) # this is bad !
			############## output
			return g_ix, g_iθ
		self.model = gradiator_2(fwd,bwd)
	def forward(self,x): return self.model(x,self.θ)

for N in [4,8]:
	identifier = f"fc({N})-LinearGxNoiseMul"
	classifier = nn.Sequential(
		nn.Flatten(),
		*nn_fc(expspace(28*28,10,N+1),nn.GELU(),dropout=0,batchnorm=1,gen_linear=lambda dim_i,dim_o:LinearXθNoiseMul(dim_i,dim_o,True,False)),
		nn.Softmax(dim=1)
	)
	analysis_classification(classifier, identifier, nn.MSELoss(),)
	identifier = f"fc({N})-LinearGθNoiseMul"
	classifier = nn.Sequential(
		nn.Flatten(),
		*nn_fc(expspace(28*28,10,N+1),nn.GELU(),dropout=0,batchnorm=1,gen_linear=lambda dim_i,dim_o:LinearXθNoiseMul(dim_i,dim_o,False,True)),
		nn.Softmax(dim=1)
	)
	analysis_classification(classifier, identifier, nn.MSELoss(),)

################################################################
################################################################
################################################################
################################################################
################################################################
################################################################
################################################################
################################################################
################################################################
################################################################
################################################################

class LinearXθNoiseAddBig(nn.Module):
	def __init__(self,dim_i,dim_o,gx_noise,gθ_noise):
		super().__init__()
		self.θ = nn.Parameter(nn.init.xavier_normal_(torch.empty(dim_i, dim_o)))
		self.b = nn.Parameter(torch.randn(dim_o))
		def fwd(x,θ): return torch.einsum('ij,ni->nj', θ, x) + self.b
		def bwd(x,θ,y,g_o,debug=False):
			############## normal stuff
			g_iθ = torch.einsum('ni,nj->nij', x, g_o) # produit extérieur
			g_ix = torch.einsum('nj,ij->ni' , g_o, θ) # θ⋅x
			############## alter
			if gx_noise: g_ix = g_ix + 5*torch.randn(g_ix.shape,device=g_ix.device) # this seems fine...
			if gθ_noise: g_iθ = g_iθ + 5*torch.randn(g_iθ.shape,device=g_iθ.device) # this is bad !
			############## output
			return g_ix, g_iθ
		self.model = gradiator_2(fwd,bwd)
	def forward(self,x): return self.model(x,self.θ)

for N in [4,8]:
	identifier = f"fc({N})-LinearGxNoiseAddBig"
	classifier = nn.Sequential(
		nn.Flatten(),
		*nn_fc(expspace(28*28,10,N+1),nn.GELU(),dropout=0,batchnorm=1,gen_linear=lambda dim_i,dim_o:LinearXθNoiseAddBig(dim_i,dim_o,True,False)),
		nn.Softmax(dim=1)
	)
	analysis_classification(classifier, identifier, nn.MSELoss(),)
	identifier = f"fc({N})-LinearGθNoiseAddBig"
	classifier = nn.Sequential(
		nn.Flatten(),
		*nn_fc(expspace(28*28,10,N+1),nn.GELU(),dropout=0,batchnorm=1,gen_linear=lambda dim_i,dim_o:LinearXθNoiseAddBig(dim_i,dim_o,False,True)),
		nn.Softmax(dim=1)
	)
	analysis_classification(classifier, identifier, nn.MSELoss(),)

################################################################
################################################################
################################################################
################################################################
################################################################
################################################################

class LinearXθNoiseMulBig(nn.Module):
	def __init__(self,dim_i,dim_o,gx_noise,gθ_noise):
		super().__init__()
		self.θ = nn.Parameter(nn.init.xavier_normal_(torch.empty(dim_i, dim_o)))
		self.b = nn.Parameter(torch.randn(dim_o))
		def fwd(x,θ): return torch.einsum('ij,ni->nj', θ, x) + self.b
		def bwd(x,θ,y,g_o,debug=False):
			############## normal stuff
			g_iθ = torch.einsum('ni,nj->nij', x, g_o) # produit extérieur
			g_ix = torch.einsum('nj,ij->ni' , g_o, θ) # θ⋅x
			############## alter
			if gx_noise: g_ix = g_ix * ( torch.ones(g_ix.shape,device=g_ix.device) + 5 * torch.randn(g_ix.shape,device=g_ix.device)) # this seems fine...
			if gθ_noise: g_iθ = g_iθ * ( torch.ones(g_iθ.shape,device=g_iθ.device) + 5 * torch.randn(g_iθ.shape,device=g_iθ.device)) # this is bad !
			############## output
			return g_ix, g_iθ
		self.model = gradiator_2(fwd,bwd)
	def forward(self,x): return self.model(x,self.θ)

for N in [4,8]:
	identifier = f"fc({N})-LinearGxNoiseMulBig"
	classifier = nn.Sequential(
		nn.Flatten(),
		*nn_fc(expspace(28*28,10,N+1),nn.GELU(),dropout=0,batchnorm=1,gen_linear=lambda dim_i,dim_o:LinearXθNoiseMulBig(dim_i,dim_o,True,False)),
		nn.Softmax(dim=1)
	)
	analysis_classification(classifier, identifier, nn.MSELoss(),)
	identifier = f"fc({N})-LinearGθNoiseMulBig"
	classifier = nn.Sequential(
		nn.Flatten(),
		*nn_fc(expspace(28*28,10,N+1),nn.GELU(),dropout=0,batchnorm=1,gen_linear=lambda dim_i,dim_o:LinearXθNoiseMulBig(dim_i,dim_o,False,True)),
		nn.Softmax(dim=1)
	)
	analysis_classification(classifier, identifier, nn.MSELoss(),)

################################################################
################################################################
################################################################
################################################################
################################################################
################################################################
################################################################
################################################################
################################################################
################################################################
################################################################

class LinearRndOriented(nn.Module):
	def __init__(self,dim_i,dim_o):
		super().__init__()
		self.θ = nn.Parameter(nn.init.xavier_normal_(torch.empty(dim_i, dim_o)))
		self.b = nn.Parameter(torch.randn(dim_o))
		def fwd(x,θ): return torch.einsum('ij,ni->nj', θ, x) + self.b
		def bwd(x,θ,y,g_o,debug=False):
			############## normal stuff
			t_gθ = torch.einsum('ni,nj->nij', x, g_o) # produit extérieur
			t_gx = torch.einsum('nj,ij->ni' , g_o, θ) # θ⋅x
			############## inventer un truc
			gθ = torch.randn(t_gθ.shape,device=t_gθ.device)
			gx = torch.randn(t_gx.shape,device=t_gx.device)
			# print(t_gθ.shape,t_gx.shape)
			# print(gθ.shape,gx.shape)
			# return t_gx,t_gθ
			sgnθ = (t_gθ*gθ).sum(axis=[1,2]).sign().view(-1,1,1)
			sgnx = (t_gx*gx).sum(axis=[1]).sign().view(-1,1)
			gθ = sgnθ*gθ # flip direction to be correct
			gx = sgnx*gx # flip direction to be correct
			############## output
			return gx, gθ
		self.model = gradiator_2(fwd,bwd)
	def forward(self,x): return self.model(x,self.θ)

for N in [4,8]:
	identifier = f"fc({N})-LinearRndOriented"
	classifier = nn.Sequential(
		nn.Flatten(),
		*nn_fc(expspace(28*28,10,N+1),nn.GELU(),dropout=0,batchnorm=1,gen_linear=lambda dim_i,dim_o:LinearRndOriented(dim_i,dim_o)),
		nn.Softmax(dim=1)
	)
	analysis_classification(classifier, identifier, nn.MSELoss(),)

################################################################
################################################################
################################################################
################################################################
################################################################
class LinearRndOriented2(nn.Module):
	def __init__(self,dim_i,dim_o):
		super().__init__()
		self.θ = nn.Parameter(nn.init.xavier_normal_(torch.empty(dim_i, dim_o)))
		self.b = nn.Parameter(torch.randn(dim_o))
		def fwd(x,θ): return torch.einsum('ij,ni->nj', θ, x) + self.b
		def bwd(x,θ,y,g_o,debug=False):
			############## normal stuff
			t_gθ = torch.einsum('ni,nj->nij', x, g_o) # produit extérieur
			t_gx = torch.einsum('nj,ij->ni' , g_o, θ) # θ⋅x
			############## inventer un truc
			gθ = torch.randn(t_gθ.shape,device=t_gθ.device)
			gx = torch.randn(t_gx.shape,device=t_gx.device)
			# print(t_gθ.shape,t_gx.shape)
			# print(gθ.shape,gx.shape)
			# return t_gx,t_gθ
			sgnθ = (t_gθ*gθ).sign()
			sgnx = (t_gx*gx).sign()
			gθ = sgnθ*gθ # flip direction to be correct
			gx = sgnx*gx # flip direction to be correct
			############## output
			return gx, gθ
		self.model = gradiator_2(fwd,bwd)
	def forward(self,x): return self.model(x,self.θ)

for N in [4,8]:
	identifier = f"fc({N})-LinearRndOriented2"
	classifier = nn.Sequential(
		nn.Flatten(),
		*nn_fc(expspace(28*28,10,N+1),nn.GELU(),dropout=0,batchnorm=1,gen_linear=lambda dim_i,dim_o:LinearRndOriented2(dim_i,dim_o)),
		nn.Softmax(dim=1)
	)
	analysis_classification(classifier, identifier, nn.MSELoss(),)

################################################################
################################################################
################################################################
################################################################
################################################################

class LinearBroken(nn.Module):
	def __init__(self,dim_i,dim_o,
		θ_noise=0.0,
		θ_only_sign=False,
		# θ_flip=0.0, # 0.3,
		gθ_flip=0.0, # 0.48
		gx_flip=0.0, # 0.48
		gθ_only_sign=False,
		gx_only_sign=False,
	):
		super().__init__()
		self.θ = nn.Parameter(nn.init.xavier_normal_(torch.empty(dim_i, dim_o)))
		self.b = nn.Parameter(torch.randn(dim_o))
		def fwd(x,θ): return torch.einsum('ij,ni->nj', θ, x) + self.b
		def bwd(x,θ,y,g_o,debug=False):
			############## break θ
			if not θ_noise==0 : θ = θ + θ_noise*torch.randn(dim_i,dim_o)
			if θ_only_sign: θ = only_sign(θ)
			# θ = random_sign_flip(θ, θ_flip)
			############## normal stuff
			g_iθ = torch.einsum('ni,nj->nij', x, g_o) # produit extérieur
			g_ix = torch.einsum('nj,ij->ni' , g_o, θ) # θ⋅x
			############## flip
			g_iθ = g_iθ * self.gθ_flip_static
			g_ix = g_ix * self.gx_flip_static
			if not gx_flip_dynamic==0.0 : g_ix = random_sign_flip(g_ix, gx_flip_dynamic)  # this is
			if not gθ_flip_dynamic==0.0 : g_iθ = random_sign_flip(g_iθ, gθ_flip_dynamic)  # but this is ok ... ??
			############## only sign
			if gx_only_sign: g_ix = only_sign(g_ix) # this seems fine...
			if gθ_only_sign: g_iθ = only_sign(g_iθ) # this is bad !
			############## output
			return g_ix, g_iθ
		self.model = gradiator_2(fwd,bwd)
	def forward(self,x): return self.model(x,self.θ)


def helper(
	N=4,
	θ_noise_static=0.0,
	θ_noise_dyn   =0.0,
	θ_only_sign   =False,
	θ_flip        =0.0, # 0.3,
	gθ_flip_static=0.0,
	gx_flip_static=0.0,
	# gθ_noise_dyn_σ=0.0,
	# gx_noise_dyn_σ=0.0,
	gθ_flip_dynamic=0.0, # 0.48
	gx_flip_dynamic=0.0, # 0.48
	gθ_only_sign=False,
	gx_only_sign=False,
):
	identifier = f"fc({N})-θnoise=[{θ_noise_static},{θ_noise_dyn}]-gθflip=[{gθ_flip_static},{gθ_flip_dynamic}]"
	identifier = f"{identifier}-gxflip=[{gx_flip_static},{gx_flip_dynamic}]-sgn_only=[{gx_only_sign},{gθ_only_sign}]"
	classifier = nn.Sequential(
		nn.Flatten(),
		*nn_fc(expspace(28*28,10,N+1),aa.modulize(lambda x:x**2),dropout=0,batchnorm=1,gen_linear=lambda dim_i,dim_o:LinearBroken(dim_i,dim_o,
			θ_noise_static=0.0,
			θ_noise_dyn   =0.0,
			θ_only_sign   =False,
			θ_flip        =0.0, # 0.3,
			gθ_flip_static=0.0,
			gx_flip_static=0.0,
			# gθ_noise_dyn_σ=0.0,
			# gx_noise_dyn_σ=0.0,
			gθ_flip_dynamic=0.0, # 0.48
			gx_flip_dynamic=0.0, # 0.48
			gθ_only_sign=False,
			gx_only_sign=False,
		)),
		nn.Softmax(dim=1)
	)
	analysis_classification(classifier, identifier, nn.MSELoss(),)






helper(
	N=4,
	θ_noise_static=0.0,
	θ_noise_dyn   =0.0,
	θ_only_sign   =False,
	θ_flip        =0.0, # 0.3,
	gθ_flip_static=0.0,
	gx_flip_static=0.0,
	# gθ_noise_dyn_σ=0.0,
	# gx_noise_dyn_σ=0.0,
	gθ_flip_dynamic=0.0, # 0.48
	gx_flip_dynamic=0.0, # 0.48
	gθ_only_sign=False,
	gx_only_sign=False,
)

################################################################
################################################################
################################################################
################################################################
################################################################
################################################################
################################################################

class LinearBrokenNoise(nn.Module):
	def __init__(self,dim_i,dim_o,
		gx_noise=0.0,
		gθ_noise=0.0,
	):
		super().__init__()
		self.θ = nn.Parameter(nn.init.xavier_normal_(torch.empty(dim_i, dim_o)))
		self.b = nn.Parameter(torch.randn(dim_o))
		def fwd(x,θ,b): return torch.einsum('ij,ni->nj', θ, x) + b
		def bwd(x,θ,b,y,go,debug=False):
			gx = torch.einsum('nj,ij->ni' , go, θ) # θ⋅x
			gθ = torch.einsum('ni,nj->nij', x, go) # produit extérieur
			gb = go                                # produit extérieur
			# print("gx",torch.flatten(gx).abs().mean())
			# print("gθ",torch.flatten(gθ).abs().mean())
			# print("gb",torch.flatten(gb).abs().mean())
			############## destroy
			gx = gx + torch.flatten(gx).abs().mean() * gx_noise * torch.randn(gx.shape,device=gx.device)
			gθ = gθ + torch.flatten(gθ).abs().mean() * gθ_noise * torch.randn(gθ.shape,device=gθ.device)
			############## output
			return gx, gθ, gb
		self.model = gradiator_3(fwd,bwd)
	def forward(self,x): return self.model(x,self.θ,self.b)


def helper(
	N=4,
	gx_noise=3.0,
	gθ_noise=3.0,
):
	identifier = f"fc({N})-gxnoise={gx_noise}-gθnoise={gθ_noise}"
	classifier = nn.Sequential(
		nn.Flatten(),
		*nn_fc(expspace(28*28,10,N+1),aa.modulize(lambda x:x**2),dropout=0,batchnorm=1,gen_linear=lambda dim_i,dim_o:LinearBrokenNoise(dim_i,dim_o,
			gx_noise=gx_noise,
			gθ_noise=gθ_noise,
		)),
		nn.Softmax(dim=1)
	)
	analysis_classification(classifier, identifier, nn.MSELoss(),perf_batch_size=2048, epochs=32)

helper(N=4,gx_noise=10.,gθ_noise=10.,)
helper(N=4,gx_noise=0.0,gθ_noise=0.0,)

################################################################
################################################################
################################################################
################################################################
################################################################
################################################################
################################################################

class LinearBrokenNoise2(nn.Module):
	def __init__(self,dim_i,dim_o,
		gx_noise=0.0,
		gθ_noise=0.0,
		mean_offset=0.0,
	):
		super().__init__()
		self.θ = nn.Parameter(nn.init.xavier_normal_(torch.empty(dim_i, dim_o)))
		self.b = nn.Parameter(torch.randn(dim_o))
		def fwd(x,θ,b): return torch.einsum('ij,ni->nj', θ, x) + b
		def bwd(x,θ,b,y,go,debug=False):
			gx = torch.einsum('nj,ij->ni' , go, θ) # θ⋅x
			gθ = torch.einsum('ni,nj->nij', x, go) # produit extérieur
			gb = go                                # produit extérieur
			############## destroy
			mean_gx = mean_offset*(-1+2*(torch.randn(gx.shape,device=gx.device)>0))
			mean_gθ = mean_offset*(-1+2*(torch.randn(gθ.shape,device=gθ.device)>0))
			gx = gx * (1 + mean_gx + gx_noise * torch.randn(gx.shape,device=gx.device))
			gθ = gθ * (1 + mean_gθ + gθ_noise * torch.randn(gθ.shape,device=gθ.device))
			############## output
			return gx, gθ, gb
		self.model = gradiator_3(fwd,bwd)
	def forward(self,x): return self.model(x,self.θ,self.b)




x_noise=5.0
mean_offset=10.0
x = torch.ones(2**18)
mean_x = mean_offset*(-1+2*(torch.randn(x.shape)>0))
x = x * (1 + mean_x + x_noise * torch.randn(x.shape))
plt.clf()
plt.close("all");plt.figure(figsize=[6.4,2.4])
plt.hist(x,256,density=True)
# plt.show()
plt.savefig("out/gxnoise_mean_preview.png")























def helper(
	N=4,
	gx_noise=3.0,
	gθ_noise=3.0,
	mean_offset=0.0,
):
	identifier = f"fc({N})-gxnoise={gx_noise}-gθnoise={gθ_noise}-mean_offset={mean_offset}"
	classifier = nn.Sequential(
		nn.Flatten(),
		*nn_fc(expspace(28*28,10,N+1),aa.modulize(lambda x:x**2),dropout=0,batchnorm=1,gen_linear=lambda dim_i,dim_o:LinearBrokenNoise2(dim_i,dim_o,
			gx_noise=gx_noise,
			gθ_noise=gθ_noise,
			mean_offset=mean_offset,
		)),
		nn.Softmax(dim=1)
	)
	analysis_classification(classifier, identifier, nn.MSELoss(),perf_batch_size=2048, epochs=32)

helper(N=4,gx_noise=0.0,gθ_noise=0.0,mean_offset=5.0,)
helper(N=4,gx_noise=0.0,gθ_noise=0.0,mean_offset=10.0,)
helper(N=4,gx_noise=5.0,gθ_noise=5.0,mean_offset=10.0,)

helper(N=4,gx_noise=10.0,gθ_noise=10.0,mean_offset=20.0,)
helper(N=4,gx_noise=10.0,gθ_noise=10.0,mean_offset=10.0,)

# helper(N=4,gx_noise=5.,gθ_noise=5.,mean_offset=1.0,)
# helper(N=4,gx_noise=10.,gθ_noise=10.,)

################################################################
################################################################
################################################################
################################################################
################################################################
################################################################
################################################################
################################################################
################################################################
################################################################
################################################################
################################################################
################################################################
################################################################
################################################################
################################################################
################################################################
################################################################
################################################################
################################################################
################################################################
################################################################

def blur(x,N=1,k=8):
	hat = 1-np.abs(np.linspace(-1,1,k+1))
	hat = hat / sum(hat)
	out = x.cpu().numpy()
	for n in range(N):
		out = np.convolve(out,hat,'same')
		out[-1]=x[-1]
		out[0]=x[0]
	return torch.tensor(out)

def cantor_mad():
	if not hasattr(cantor_mad, 'array'): cantor_mad.array = aa.cantor_function(torch.linspace(0,1,2**12),5).to(device)
	return aa.modulize(gradiator_1(
		lambda x: -x + 2*(torch.floor(x) + aa.sampler1d_repeat(cantor_mad.array,x,aa.nearest_floor)),
		lambda x,y,grad: -torch.ones(x.shape,device=x.device) # derivative is 0 everywhere
	))

def cantor_mad_cheat():
	if not hasattr(cantor_mad,       'array'): cantor_mad.array       = aa.cantor_function(torch.linspace(0,1,2**12),5).to(device)
	if not hasattr(cantor_mad_cheat, 'array'):
		cantblur = aa.cantor_function(torch.linspace(0,1,3**5+1),5)
		cantblur = blur(cantblur)
		yy = torch.concat([cantblur,1+cantblur[:1]])
		diff = yy[1:] - yy[:-1]
		cantor_mad_cheat.array = cantblur.to(device)
	return aa.modulize(gradiator_1(
		lambda x: -x + 2*(torch.floor(x) + aa.sampler1d_repeat(cantor_mad.array,x,aa.nearest_floor)),
		lambda x,y,grad: +torch.ones(x.shape,device=x.device) # derivative is 0 everywhere
	))


import matplotlib
matplotlib.use('qtagg')
# cantblur = aa.cantor_function(torch.linspace(0,1,3**5+1),5)
x = torch.linspace(-2,2,2**15).to(device)
cantblur = cantor_mad()(x).detach().cpu()
x = x.cpu()
res = 2**12
yy = cantblur
yy = blur(yy,N=256,k=64)
# plt.clf()
# plt.plot(x,yy)
# plt.show()

diff = yy[1:] - yy[:-1]
diff = diff[len(diff)//2:(3*len(diff))//4] *(2**13)
plt.clf()
plt.plot(torch.linspace(0,1,len(diff)),diff)
plt.plot(x,yy)
# plt.plot(yy[len(yy)//2:(3*len(yy))//4])
plt.show()
cantor_mad_cheat.array = cantblur.to(device)




# def cantor_mad():
# 	if not hasattr(cantor_mad, 'array'): cantor_mad.array = aa.cantor_function(torch.linspace(0,1,3**5+1),5).to(device)
# 	return aa.modulize(gradiator_1(
# 		lambda x: -x + 2*(.5+torch.floor(x-.5) + aa.sampler1d_repeat(cantor_mad.array,x-.5,aa.nearest_floor)),
# 		lambda x,y,grad: -torch.ones(x.shape,device=x.device) # derivative is 0 everywhere
# 	))

# def cantor_mad_cheat():
# 	if not hasattr(cantor_mad, 'array'): cantor_mad.array = aa.cantor_function(torch.linspace(0,1,3**5+1),5).to(device)
# 	return aa.modulize(gradiator_1(
# 		lambda x: -x + 2*(.5+torch.floor(x-.5) + aa.sampler1d_repeat(cantor_mad.array,x-.5,aa.nearest_floor)),
# 		lambda x,y,grad: +torch.ones(x.shape,device=x.device) # derivative is 0 everywhere
# 	))

x = torch.linspace(-.999,1,2**15).to(device)
plt.clf()
plt.plot(x.cpu(),cantor_mad()(x).cpu())
plt.plot(x.cpu(),x.cpu())
plt.savefig("out/cantor_mad_preview.png")

DEBUG=10000

# for N in [4,8]:
for N in [4]:
# for N in [8]:
	identifier = f"fc({N})-cantor_mad-normal"
	classifier = nn.Sequential(
		nn.Flatten(),
		*nn_fc(expspace(28*28,10,N+1),cantor_mad(),dropout=0,batchnorm=1),
		nn.Softmax(dim=1)
	)
	analysis_classification(classifier, identifier, nn.MSELoss(),epochs=32,batch_size=2**8,perf_batch_size=2**14,mini_data=DEBUG)

# for N in [4,8]:
for N in [4]:
# for N in [8]:
	identifier = f"fc({N})-cantor_mad-cheat"
	classifier = nn.Sequential(
		nn.Flatten(),
		*nn_fc(expspace(28*28,10,N+1),cantor_mad_cheat(),dropout=0,batchnorm=1),
		nn.Softmax(dim=1)
	)
	analysis_classification(classifier, identifier, nn.MSELoss(),epochs=32,batch_size=2**8,perf_batch_size=2**14,mini_data=DEBUG)



################################################################
################################################################
################################################################
################################################################
################################################################

def cantor():
	if not hasattr(cantor, 'array'): cantor.array = aa.cantor_function(torch.linspace(0,1,3**5+1),5).to(device)
	return aa.modulize(gradiator_1(
		lambda x: torch.floor(x) + aa.sampler1d_repeat(cantor.array,x,aa.nearest_floor),
		# lambda x: aa.sampler1d_clamp(cantor.array,x,aa.nearest_floor),
		lambda x,y,grad: torch.zeros(x.shape,device=x.device) # derivative is 0 everywhere
	))

def cantor_fakegrad():
	if not hasattr(cantor, 'array'): cantor.array = aa.cantor_function(torch.linspace(0,1,3**5+1),5).to(device)
	return aa.modulize(gradiator_1(
		lambda x: torch.floor(x) + aa.sampler1d_repeat(cantor.array,x,aa.nearest_floor),
		lambda x,y,grad: grad # fake the gradient to identity
	))

for N in [4,8]:
	identifier = f"fc({N})-id"
	classifier = nn.Sequential(
		nn.Flatten(),
		*nn_fc(expspace(28*28,10,N+1),nn.Identity(),dropout=0,batchnorm=1),
		nn.Softmax(dim=1)
	)
	analysis_classification(classifier, identifier, nn.MSELoss(),)

for N in [4,8]:
	identifier = f"fc({N})-cantor-normal"
	classifier = nn.Sequential(
		nn.Flatten(),
		*nn_fc(expspace(28*28,10,N+1),cantor(),dropout=0,batchnorm=1),
		nn.Softmax(dim=1)
	)
	analysis_classification(classifier, identifier, nn.MSELoss(),)

for N in [4,8]:
	identifier = f"fc({N})-cantor-cheat"
	classifier = nn.Sequential(
		nn.Flatten(),
		*nn_fc(expspace(28*28,10,N+1),cantor_fakegrad(),dropout=0,batchnorm=1),
		nn.Softmax(dim=1)
	)
	analysis_classification(classifier, identifier, nn.MSELoss(),)

x = torch.linspace(-0.999,1,2**15).to(device)
plt.clf()
plt.plot(x.cpu(),cantor()(x).cpu())
plt.plot(x.cpu(),x.cpu())
plt.savefig("out/cantor_preview.png")

################################################################
################################################################
################################################################
################################################################
################################################################

ω = torch.pi/4

def square_fakegrad(step=2):
	return aa.modulize(gradiator_1(
		lambda x: torch.round(step*torch.sin(ω*x))/step,
		# lambda x: (step*torch.sin(ω*x))/step,
		lambda x,y,grad: torch.cos(ω*x)*grad # gradient forgets about rounding
	))

def square_nofakegrad(step=2):
	return aa.modulize(lambda x: torch.round(step*torch.sin(ω*x))/step)


def square_debug(step=2):
	return aa.modulize(lambda x: (step*torch.sin(ω*x))/step)


for N in [4,8]:
	identifier = f"fc({N})-square-cheat"
	classifier = nn.Sequential(
		nn.Flatten(),
		*nn_fc(expspace(28*28,10,N+1),square_fakegrad(),dropout=0,batchnorm=1),
		nn.Softmax(dim=1)
	)
	analysis_classification(classifier, identifier, nn.MSELoss(),perf_batch_size=2048)

for N in [4,8]:
	identifier = f"fc({N})-square-smooth"
	classifier = nn.Sequential(
		nn.Flatten(),
		*nn_fc(expspace(28*28,10,N+1),square_debug(),dropout=0,batchnorm=1),
		nn.Softmax(dim=1)
	)
	analysis_classification(classifier, identifier, nn.MSELoss(),perf_batch_size=2048)

for N in [4,8]:
	identifier = f"fc({N})-square-normal"
	classifier = nn.Sequential(
		nn.Flatten(),
		# *nn_fc(expspace(28*28,10,N+1),aa.modulize(lambda x:torch.round(torch.sin(x))),dropout=0,batchnorm=1),
		*nn_fc(expspace(28*28,10,N+1),square_nofakegrad(),dropout=0,batchnorm=1),
		nn.Softmax(dim=1)
	)
	analysis_classification(classifier, identifier, nn.MSELoss(),perf_batch_size=2048)

x = torch.linspace(-6,6,2**15).to(device)
plt.clf()
plt.plot(x.cpu(),square_fakegrad()(x).cpu())
plt.savefig("out/square_preview.png")
plt.plot(x.cpu(),square_debug()(x).cpu())
plt.savefig("out/square_smooth_preview.png")



################################################################
################################################################
################################################################
################################################################
################################################################

# def fixed_noise(res=32,slowdown=1):
# 	torch.manual_seed(0)
# 	# array = .5*torch.randn(res).to(device)
# 	# return aa.modulize(lambda x: (aa.sampler1d_repeat(array,x/slowdown,aa.smoothstep) - array[0]))
# 	array = .2*torch.randn(res).to(device)
# 	return aa.modulize(lambda x: aa.sampler1d_repeat(array,x/slowdown,aa.lerp))


# y = .2*torch.randn(res).to(device)
# yy = torch.concat([y,y[:1]])
# z = yy[1:] - yy[:-1]
# y = torch.tensor([0,0,0,1,1])

def blur(x,mask,N=1,k=64):
	hat = 1-np.abs(np.linspace(-1,1,k+1))
	hat = hat / sum(hat)
	out = x.cpu().numpy()
	for n in range(N):
		out = np.convolve(out,hat,'same')
		out[-k:-1]=x[-k:-1]
		out[0:k]=x[0:k]
	out = x*mask + out*(1-mask)
	return torch.tensor(out)

def resize(x,res):
	xx = torch.linspace(0,1,res)
	return aa.sampler1d_repeat(x,xx,aa.nearest_floor)


blur(x,mask,N=5,k=8)



torch.manual_seed(0)
res1=4
res2=2048
y = .2*torch.randn(res1)
plt.clf()
plt.plot(torch.linspace(0,1,res2),resize(y,res2))
plt.plot(torch.linspace(0,1,res1),y)
plt.show()



def fixed_noise(res=32,slowdown=1):
	torch.manual_seed(0)
	y = .2*torch.randn(res).to(device)
	yy = torch.concat([y,y[:1]])
	diff = yy[1:] - yy[:-1]
	return aa.modulize(gradiator_1(
		lambda x       : aa.sampler1d_repeat(y   ,x/slowdown,aa.lerp),
		lambda x,y,grad: aa.sampler1d_repeat(diff,x/slowdown,aa.nearest_floor)*grad,
	))

def fixed_noise_debug(res=32,slowdown=1):
	torch.manual_seed(0)
	y = .2*torch.randn(res)
	yy = torch.concat([y,y[:1]])
	diff = yy[1:] - yy[:-1]
	import matplotlib
	matplotlib.use('qtagg')
	plt.clf()
	x = torch.linspace(-2,2,2**15)
	plt.plot(x,aa.sampler1d_repeat(y   ,x/slowdown,aa.lerp))
	plt.plot(x,aa.sampler1d_repeat(diff,x/slowdown,aa.nearest_floor))
	plt.show()



x = torch.linspace(-2,2,2**15).to(device)
plt.clf()
# for res in [4,32,64]: plt.plot(x.cpu(),fixed_noise(res)(x).cpu())
res=4
plt.plot(x.cpu(),fixed_noise(res)(x).cpu())

plt.savefig("out/fixed_noise_preview.png")

################ !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
DEBUGGING = True
################ !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# for N in [4,8]:
# 	for res in [4,8]:

for N in [4]:
	for res in [4]:
	# for res in [32]:
	# for res in [64]:
		identifier = f"fc({N})-fixed_noise{res}"
		classifier = nn.Sequential(
			nn.Flatten(),
			*nn_fc(expspace(28*28,10,N+1),fixed_noise(),dropout=0,batchnorm=1),
			nn.Softmax(dim=1)
		)
		analysis_classification(classifier, identifier, nn.MSELoss(),mini_data=DEBUGGING)

################################################################
################################################################
################################################################
################################################################
################################################################

# def blur(x,mask=None,N=1,k=64):
# 	hat = 1-np.abs(np.linspace(-1,1,k+1))
# 	hat = hat / sum(hat)
# 	out = x
# 	for n in range(N):
# 		# print("A", out.shape)
# 		# out = np.convolve(out,hat)
# 		# out = out[hat.shape[0]//2:]
# 		# out = out[:-(hat.shape[0] - (hat.shape[0]//2) - 1)]
# 		# print("B", out.shape)
# 		# 
# 		out = np.convolve(out,hat,'same')
# 		# out[-k:-1]=x[-k:-1]
# 		# out[0:k]=x[0:k]
# 	if mask == None :
# 		mask = np.zeros(x.shape)
# 		mask[0]=mask[-1]=1
# 	# 	# mask[ :k]=1
# 	# 	# mask[-k:]=1
# 		out = x*mask + out*(1-mask)
# 	# if mask: out = x*mask + out*(1-mask)
# 	return torch.tensor(out)

# def bezos1d(xmax=10,bluriter=0):
# 	from PIL import Image, ImageOps
# 	im = Image.open("bezos-front.png")
# 	im = ImageOps.grayscale(im)
# 	im = np.array(im).T
# 	im = (im-im.min(axis=0))/(im.max(axis=0)-im.min(axis=0)) # normalize
# 	im = im*2-1 # RANGE -1 to 1
# 	im = im[:,1]
# 	im[0]=im[-1]=0 # FORCE BORDER TO 0
# 	im = blur(im,N=bluriter,k=8)
# 	y = torch.tensor(im).float().to(device)
# 	res = im.shape[0]
# 	###
# 	# res = 2048
# 	# y = torch.relu(torch.linspace(-xmax,xmax,res)).to(device)
# 	###
# 	yy = torch.concat([y,y[:1]])
# 	diff = yy[1:] - yy[:-1]
# 	# return aa.modulize(gradiator_1(
# 	# 	lambda x       : aa.sampler1d_repeat(y   ,x/slowdown,aa.lerp),
# 	# 	lambda x,y,grad: aa.sampler1d_repeat(diff,x/slowdown,aa.nearest_floor)*grad,
# 	# ))
# 	###
# 	return aa.modulize(gradiator_1(
# 		lambda x       : aa.sampler1d_clamp(y   ,x/(2*xmax)+.5),
# 		lambda x,y,grad: aa.sampler1d_clamp(diff,x/(2*xmax)+.5)*grad,
# 		# lambda x       : y[(x*res/(2*xmax)+xmax).int()],
# 		# lambda x,y,grad: (x>0)*grad,
# 	))
# 	###
# 	# return aa.modulize(gradiator_1(
# 	# 	lambda x       : torch.maximum(torch.tensor(0),x),
# 	# 	lambda x,y,grad: (x>0)*grad,
# 	# ))
# 	###
# 	# def fwd(x):
# 	# 	print("fwd", x.flatten().shape)
# 	# 	print("fwd", x.flatten().abs().mean())
# 	# 	return torch.maximum(torch.tensor(0),x)
# 	# def bwd(x,y,grad):
# 	# 	return (x>0)*grad
# 	# return aa.modulize(gradiator_1(fwd,bwd,))



def blur(x,mask=None,N=1,k=64):
	hat = 1-np.abs(np.linspace(-1,1,k+1))
	hat = hat / sum(hat)
	out = x
	for n in range(N): out = np.convolve(out,hat,'same')
	if mask == None :
		mask = np.zeros(x.shape)
		mask[0]=mask[-1]=1
		out = x*mask + out*(1-mask)
	# if mask: out = x*mask + out*(1-mask)
	return torch.tensor(out)

# def bezos1d(xmax=10,bluriter=0):
# 	from PIL import Image, ImageOps
# 	im = Image.open("bezos-front.png")
# 	im = ImageOps.grayscale(im)
# 	im = np.array(im).T
# 	im = (im-im.min(axis=0))/(im.max(axis=0)-im.min(axis=0)) # normalize
# 	im = im*2-1 # RANGE -1 to 1
# 	im = im[:,1]
# 	im[0]=im[-1]=0 # FORCE BORDER TO 0
# 	im = blur(im,N=bluriter,k=8)
# 	y = torch.tensor(im).float().to(device)
# 	res = im.shape[0]
# 	###
# 	yy = torch.concat([y,y[:1]])
# 	diff = yy[1:] - yy[:-1]
# 	###
# 	return aa.modulize(gradiator_1(
# 		lambda x       : aa.sampler1d_clamp(y   ,x/(2*xmax)+.5),
# 		lambda x,y,grad: aa.sampler1d_clamp(diff,x/(2*xmax)+.5)*grad,
# 	))

def bezos1d(xmax=10,bluriter=0,blur_only_bwd=False):
	from PIL import Image, ImageOps
	im = Image.open("bezos-front.png")
	im = ImageOps.grayscale(im)
	im = np.array(im).T
	im = (im-im.min(axis=0))/(im.max(axis=0)-im.min(axis=0)) # normalize
	im = im*2-1 # RANGE -1 to 1
	im = im[:,1]
	im[0]=im[-1]=0 # FORCE BORDER TO 0
	im_noblur = im
	im = blur(im,N=bluriter,k=8)
	y = torch.tensor(im).float().to(device)
	y[0]=y[1]=0
	y[-1]=y[-2]=0
	res = im.shape[0]
	###
	yy = torch.concat([y,y,y,y])
	# diff = yy[len(y)+1:2*len(y)+1] - yy[len(y)-1:2*len(y)-1] * len(y)/xmax/2
	diff = yy[len(y)+1:2*len(y)+1] - yy[len(y)-1:2*len(y)-1] * len(y)/xmax/4
	###
	if blur_only_bwd: # reset it as it was
		y = torch.tensor(im).float().to(device)
		y[0]=y[1]=0
		y[-1]=y[-2]=0
	return aa.modulize(gradiator_1(
		lambda x       : aa.sampler1d_clamp(y   ,x/(2*xmax)+.5),
		lambda x,y,grad: aa.sampler1d_clamp(diff,x/(2*xmax)+.5)*grad,
	))

# import matplotlib
# matplotlib.use('qtagg')
x = torch.linspace(-20,20,2**15).to(device)
plt.clf()
plt.close("all");plt.figure(figsize=[6.4,2.4])

for i in [0,4,8]:
	plt.plot(x.cpu(),bezos1d(bluriter=i)(x).cpu())

plt.savefig("out/bezos1d_preview.png")
# plt.show()

datasize = False
datasize = 60000
datasize = 1200 # ??? like what they do with SOLO
# DEBUGGING = 10000 # ??? like what they do with SOLO
datasize = 10000 # ??? like what they do with SOLO

for N in [4]:
	for blur_only_bwd in [False,True]:
		for bluriter in [0,2,4,8]:
		# for bluriter in [0]:
		# for bluriter in [2]:
		# for bluriter in [4]:
		# for bluriter in [8]:
			identifier = f"fc({N})-bezos1d-blur={bluriter}-datasize={datasize}-blur_only_bwd={blur_only_bwd}"
			classifier = nn.Sequential(
				nn.Flatten(),
				*nn_fc(expspace(28*28,10,N+1),bezos1d(bluriter=bluriter,blur_only_bwd=blur_only_bwd),dropout=0,batchnorm=1),
				# *nn_fc(expspace(28*28,10,N+1),nn.ReLU(),dropout=0,batchnorm=1),
				nn.Softmax(dim=1)
			)
			analysis_classification(classifier, identifier, nn.MSELoss(),batch_size=2**8,perf_batch_size=2**14,mini_data=datasize,epochs=64)


################################################################
################################################################
################################################################
################################################################
################################################################

def bezos_loader_2d(path,blur=0,minus1to1=True):
	from PIL import Image, ImageOps, ImageFilter
	im = Image.open(path)
	im = ImageOps.grayscale(im)
	im = im.filter(ImageFilter.GaussianBlur(radius=blur))
	im = np.array(im).T
	im = (im-im.min())/(im.max()-im.min()) # normalize
	if minus1to1 : im = im*2-1 # RANGE -1 to 1
	##############
	# TODO # im[0]=im[-1]=0 # FORCE BORDER TO 0 to avoir SIGMOID like utilization
	##############
	return torch.tensor(im).float().to(device)

for blur in [0,4]:
	b1=bezos_loader_2d("bezos-front2a.png",blur=blur,minus1to1=False)
	b2=bezos_loader_2d("bezos-front2b.png",blur=blur,minus1to1=False)
	y=torch.stack([b1,torch.zeros(b2.shape,device=b2.device),b2,])
	from PIL import Image
	im = (255*y).permute([2,1,0]).cpu().byte()
	Image.fromarray(im.numpy()).save(f"out/bezos2d_blur={blur}_preview.png")
	# plt.clf()
	# plt.imshow(im)
	# plt.savefig(f"out/bezos2d_blur={blur}_preview.png")
	# plt.show()

def bezos2d(xmax=10,blur=0):
	b1=bezos_loader_2d("bezos-front2a.png",blur=blur)
	b2=bezos_loader_2d("bezos-front2b.png",blur=blur)
	y = torch.stack([b1,torch.zeros(b2.shape,device=b2.device),b2,])
	###
	# y_x = torch.concat([y,y[:1]])
	# diff = yy[1:] - yy[:-1]
	###
	def fwd(x):
		print("FORWARD")
		print(x.shape)
		return aa.sampler2d_nearest_clamp(y   ,x.view(-1,2)/(2*xmax)+.5).view(x.shape)
	def bwd(x,y,grad):
		print("BACKWARD")
		return aa.sampler2d_nearest_clamp(diff,x.view(-1,2)/(2*xmax)+.5).view(x.shape)*grad,
	return aa.modulize(gradiator_1(fwd,bwd))

datasize = 1200

for N in [4]:
	# for bluriter in [0,4,8]:
	# for bluriter in [0]:
	# for bluriter in [4]:
	for blur in [8]:
		identifier = f"fc({N})-bezos2d-blur={bluriter}-datasize={datasize}"
		classifier = nn.Sequential(
			nn.Flatten(),
			*nn_fc(expspace(28*28,10,N+1),bezos2d(blur=blur),dropout=0,batchnorm=1),
			# *nn_fc(expspace(28*28,10,N+1),nn.ReLU(),dropout=0,batchnorm=1),
			nn.Softmax(dim=1)
		)
		analysis_classification(classifier, identifier, nn.MSELoss(),batch_size=2**6,perf_batch_size=2**14,mini_data=datasize)





























import matplotlib
matplotlib.use('qtagg')
x = torch.linspace(-20,20,2**15).to(device)
plt.clf()
for i in [0,2,4,8]:
	plt.plot(x.cpu(),bezos1d(bluriter=i)(x).cpu())

plt.show()

datasize = False
datasize = 60000
datasize = 1200 # ??? like what they do with SOLO
# DEBUGGING = 10000 # ??? like what they do with SOLO

for N in [4]:
	# for bluriter in [0,4,8]:
	# for bluriter in [0]:
	# for bluriter in [4]:
	for bluriter in [8]:
		identifier = f"fc({N})-bezos1d-blur={bluriter}-datasize={datasize}"
		classifier = nn.Sequential(
			nn.Flatten(),
			*nn_fc(expspace(28*28,10,N+1),bezos1d(bluriter=bluriter),dropout=0,batchnorm=1),
			# *nn_fc(expspace(28*28,10,N+1),nn.ReLU(),dropout=0,batchnorm=1),
			nn.Softmax(dim=1)
		)
		analysis_classification(classifier, identifier, nn.MSELoss(),batch_size=2**6,perf_batch_size=2**14,mini_data=datasize)


################################################################
################################################################
################################################################
################################################################
################################################################

class Random_Mapping(nn.Module):
	def __init__(self,dim_i,dim_o,b=0):
		super().__init__()
		# self.θ = nn.Parameter(nn.init.xavier_normal_(torch.empty(dim_i, dim_o)),False)
		# self.θ = nn.Parameter(torch.zeros(dim_i, dim_o).fill_diagonal_(1),False)
		self.θ = nn.Parameter(torch.randn(dim_i, dim_o),False)
		self.b = nn.Parameter(torch.randn(dim_o),False)
		# self.b = nn.Parameter(torch.zeros(dim_o),False)
		def fwd(x): return torch.einsum('ij,ni->nj', self.θ, x) + self.b
		def bwd(x,y,go,debug=False):
			gx = torch.einsum('nj,ij->ni' , go, self.θ)
			return gx
		self.model = gradiator_1(fwd,bwd)
	def forward(self,x): return self.model(x)

datasize = 60000
# datasize = 1200 # ??? like what they do with SOLO
# idėe : SOLO fait des images 128x128, mais elle ne sont pas arbitraire.
#        La véritable dimension des outputs doit être bcp plus basse
#        disons 10%, et ça laisse bcp de marge
rnd_dim = int((128*128)*0.1)
rnd_dim = int((128*128)*0.2)
identifier = f"fc-random-mapping-datasize={datasize}-rndim={rnd_dim}"
classifier = nn.Sequential(
	nn.Flatten(),
	Random_Mapping(28*28,rnd_dim),             # messup data
	nn.ReLU(),Random_Mapping(rnd_dim,rnd_dim), # messup data
	nn.ReLU(),Random_Mapping(rnd_dim,rnd_dim), # messup data
	nn.ReLU(),Random_Mapping(rnd_dim,rnd_dim), # messup data
	nn.ReLU(),Random_Mapping(rnd_dim,rnd_dim), # messup data
	nn.BatchNorm1d(rnd_dim),
	nn.Linear(rnd_dim,10),
	nn.Softmax(dim=1)
)
analysis_classification(classifier, identifier, nn.MSELoss(),batch_size=2**6,perf_batch_size=2**14,mini_data=datasize)

datasize = 60000
# datasize = 1200 # ??? like what they do with SOLO
identifier = f"fc-no-random-mapping-datasize={datasize}"
# idėe : SOLO fait des images 128x128, mais elle ne sont pas arbitraire.
#        La véritable dimension des outputs doit être bcp plus basse
#        disons 10%, et ça laisse bcp de marge
rnd_dim = int((128*128)*0.1)
classifier = nn.Sequential(
	nn.Flatten(),
	nn.Linear(28*28,10),
	nn.Softmax(dim=1)
)
analysis_classification(classifier, identifier, nn.MSELoss(),batch_size=2**6,perf_batch_size=2**14,mini_data=datasize)

























































###################################################################################
###################################################################################
###################################################################################
# L = BrokenLinear(10,20)
# x = torch.randn(5,10)
# L(x).shape

class BrokenClassifier(nn.Module):
	def __init__(self,layers,gen_linear):
		super().__init__()
		layers = [int(np.round(x)) for x in layers]
		self.L  = [gen_linear(λ1,λ2)  for (λ1,λ2) in zip(layers[:-1],layers[1:])]
		self.BN = [nn.BatchNorm1d(λ2) for (λ1,λ2) in zip(layers[:-1],layers[1:])]
		self.σ  = [nn.GELU()          for (λ1,λ2) in zip(layers[:-1],layers[1:])]
		self.model = nn.Sequential(*reduce(operator.add, zip(self.L,self.σ,self.BN))[:-2])
		self.softmax = nn.Softmax(dim=1)
	def forward(self,x):
		return self.softmax(self.model(nn.Flatten()(x)))

# layers = [28*28,32,10]
# cl = BrokenClassifier(layers)
# cl(torch.randn(10,1,28,28))

# def gen_linear(λ1,λ2): return BrokenLinear(λ1,λ2) # TODOOOOOOOOOOOOO
# layers = [28*28,32,10]
# cl = BrokenClassifier(layers,gen_linear)
# cl(torch.randn(10,1,28,28))

fmnist_L_x = mnist_analysis.fmnist_L_x
fmnist_L_y = mnist_analysis.fmnist_L_y
fmnist_T_x = mnist_analysis.fmnist_T_x
fmnist_T_y = mnist_analysis.fmnist_T_y

os.makedirs("models",exist_ok=True)

def analysis(
	N,
	θ_noise_static_σ=0.0,
	θ_noise_static_μ=0.0,
	gθ_flip_static=0.0,
	gx_flip_static=0.0,
	gθ_flip_dynamic=0.0, # 0.48
	gx_flip_dynamic=0.0, # 0.48
	gθ_only_sign=False,
	gx_only_sign=False,
):
	identifier = f"fc({N})-θnoise=[{θ_noise_static_σ},{θ_noise_static_μ}]-flip_static=[{gθ_flip_static},{gx_flip_static}]"
	identifier = f"{identifier}-flip_dyn=[{gθ_flip_dynamic},{gx_flip_dynamic}]-sgn_only=[{gx_only_sign},{gθ_only_sign}]"
	# model_path = f"models/{identifier}"
	def gen_linear(λ1,λ2):
		return BrokenLinear(λ1,λ2,
			θ_noise_static_σ = θ_noise_static_σ ,
			θ_noise_static_μ = θ_noise_static_μ ,
			gθ_flip_static   = gθ_flip_static   ,
			gx_flip_static   = gx_flip_static   ,
			gθ_flip_dynamic  = gθ_flip_dynamic  ,
			gx_flip_dynamic  = gx_flip_dynamic  ,
			gθ_only_sign     = gθ_only_sign     ,
			gx_only_sign     = gx_only_sign     ,
		)
	classifiers = BrokenClassifier(expspace(28*28,10,N+1),gen_linear)
	mnist_analysis.analysis_classification(classifiers, identifier, nn.MSELoss(),batch_size=2048,epochs=16)
	# torch.save(classifiers, model_path)

# analysis(4,0.0,0.0)
# analysis(4,0.5,0.0)
# analysis(4,20000,0.0)
analysis(4,1,0.0)
# analysis(4,0.0,200)


# for N in [2,4]:
#  for  θ_noise_static_σ in [0,0.5,1.0]
#   for θ_noise_static_μ in [0,0.5,1.0]
#    add_job(lambda:analysis(
#      N,
#      θ_noise_static_σ,
#      θ_noise_static_μ,
#    ))


# 	for gθ_flip_static   = gθ_flip_static   ,
# 	for gx_flip_static   = gx_flip_static   ,
# 	for gθ_flip_dynamic  = gθ_flip_dynamic  ,
# 	for gx_flip_dynamic  = gx_flip_dynamic  ,
# 	for gθ_only_sign     = gθ_only_sign     ,
# 	for gx_only_sign     = gx_only_sign     ,

# 								add_job(lambda:analysis(
# 									N,
# 									θ_noise_static_σ,
# 									θ_noise_static_μ,
# 									gθ_flip_static,
# 									gx_flip_static,
# 									gθ_flip_dynamic,
# 									gx_flip_dynamic,
# 									gθ_only_sign,
# 									gx_only_sign,
# 								))
